function [net, se] = dlptrain(net, xd, yd, opt)
% [net, se] = slptrain(net, xd, yd, opt) 
%   Funkcja uczy siec odpowiadac sygnalami yd na sygnaly wejscowe xd,
%   parametr opt decyduje o parametrach procesu uczenia
% 
% Argumenty wejsciowe:
%   net  - zmienna reprezentujaca perceptron dwuwarstwowy
%   xd   - kolumnowy wektor sygnalow wejsciowych lub 
%          macierz ktorej kolumny zawieraja kolejne wektory sygnalow wejsciowych
%   yd   - kolumnowy wektor sygnalow wyjsciowych lub 
%          macierz ktorej kolumny zawieraja kolejne wektory sygnalow wyjsciowych
%   opt  - tablica komorkowa zawierajaca opcje:
%
%     view - parametr decydujacy o pokazywaniu szczegolow przebiegu procesu
%              uczenia, dopuszczalne wartosci:
%          0 - szczegoly nie sa pokazywane
%          1 - w kazdym kroku algorytmu pokazywane sa:
%              numer iteracji oraz wartosc bezwgledna sumy bledow dla wzorcow 
%              zawartych w xd (se)
%          2 - w kazdym kroku algorytmu wywolywana jest funkcja 
%              uzytkownika, ktorej nazwe nalezy podac w opcji viewFn
%          domyslnie view = 1
%     viewFn - parametr zawierajacy nazwe funkcji uzytkownika ktora jest 
%          opowiedzialna za pokazywanie szczegolow procesu uczenia,
%          funkcja ta powinna miec nastepujacy naglowek:
%          function F(net, it, se)
%          parametry wejsciowe funkcji zawieraja:
%            net - aktualne dane perceptronu
%            it  - numer aktualnej iteracji
%            se  - wartosc bezwgledna sumy bledow dla wzorcow zawartych w xd
%     maxit - okresla maksymalna liczbe prezentacji wzorcow
%          domyslnie maxit = 1000
%     c     - wspolczynnik wykorzystywany w trakcie uczenia
%          domyslnie c = 0.1
%     eps - okresla wymagana maksymalna wartosc ktora musi osiagnac suma bledow
%           dla wzorcow wykorzystywanych w trakcie uczenia
%           domyslnie eps = 0.001
%          
% Argumenty wyjsciowe:
%   net  - zmienna reprezentujaca siec zawierajaca nowe wartosci wag i biasow
%   se   - wartosc bezwzgledna z sumy bledow dla prezentowanych wzorcow

if (strcmp(net.type, 'dlp') == 0) 
  error('Siec nie jest perceptronem wielowarstwowym\n');
end

if (size(xd,1) ~= net.nin)
  error('Nieprawidlowy rozmiar wektora sygnalow wejsciowych\n');
end

if (size(yd,1) ~= net.nout)
  error('Nieprawidlowy rozmiar wektora sygnalow wyjsciowych\n');
end

if (size(xd,2) ~= size(yd,2))
  error('Liczba wzorcow w xd nie jest rowna liczbie wzorcow w yd\n');
end

if sum(strcmp(net.hiddenfn, {'linear', 'logistic', 'tanh'})) == 0
  error('Funkcja aktywacji neuronow sieci musi byc funkcja ciagla\n');
end

if sum(strcmp(net.outfn, {'linear', 'logistic', 'tanh'})) == 0
  error('Funkcja aktywacji neuronow sieci musi byc funkcja ciagla\n');
end

if ( nargin == 3 )
  opt = {};
end

view   = findopt(opt, 'view', 1);
viewFn = findopt(opt, 'viewFn', '');
maxit  = findopt(opt, 'maxit', 1000); 
c      = findopt(opt, 'c', 0.1);      
eps    = findopt(opt, 'eps', 0.001);  

if ((view == 2) & (length(viewFn) == 0))
  error('Niezdefiniowany parametr viewFn\n');
end

if ((any(any(net.wb1)) == 0) & (any(any(net.wb2)) == 0))
  net.w1 = 1-2*rand(net.nhidden, net.nin);
  net.b1 = 1-2*rand(net.nhidden, 1);
  net.w2 = 1-2*rand(net.nout, net.nhidden);
  net.b2 = 1-2*rand(net.nout, 1); 
  net.wb1 = [net.w1 net.b1];
  net.wb2 = [net.w2 net.b2];
end

nxd = size(xd, 2); se=0;
if (~isempty(net.b1))  
  xd = [xd; ones(1, nxd)]; 
end
y11 = zeros(net.nhidden,nxd);
if (~isempty(net.b1))  
  y11 = [y11; ones(1, nxd)]; 
end

for i=1:maxit

% wyznaczanie bledu
  fi = net.wb1 * xd;
  switch net.hiddenfn
    case 'linear'    
      y1 = fi;
    case 'tanh'
      y1 = tanh(fi);
    case 'logistic'
      y1 = 1./(1+exp(-fi));
  end  
  y11 (1:net.nhidden, :) = y1;

  fi = net.wb2*y11;
  switch net.outfn
    case 'linear'    
      y2 = fi;
    case 'tanh'
      y2 = tanh(fi);
    case 'logistic'
      y2 = 1./(1+exp(-fi));
  end
  e  =  yd - y2;
  se = sum(sum(e.*e))/(2*nxd);

  switch net.outfn
    case 'linear'    
      delta2 = e;
    case 'tanh'
      delta2 = (1+y2).*(1-y2).*e;
    case 'logistic'
      delta2 = y2.*(1-y2).*e;
  end

  e1 = net.w2'*delta2;
  
  switch net.hiddenfn
    case 'linear'    
      delta1 = e1;
    case 'tanh'
      delta1 = (1+y1).*(1-y1).*e1;
    case 'logistic'
      delta1 = y1.*(1-y1).*e1;
  end  
  
% modyfikacja wag
  net.wb2 = net.wb2 + c*delta2*y11'; 
  net.wb1 = net.wb1 + c*delta1*xd'; 
  
  net.w1 = net.wb1(:, 1:net.nin);
  if (~isempty(net.b1))
    net.b1 = net.wb1(:, net.nin+1);
  end
  net.w2 = net.wb2(:, 1:net.nhidden);
  if (~isempty(net.b2))
    net.b2 = net.wb2(:, net.nhidden+1);
  end
  
  switch view
    case 1
      disp(sprintf('%5d. se = %6.4f', i, se));
    case 2
      feval(viewFn, net, i, se);      
  end 
    
  if (se < eps)
    break
  end        
end
