function [net, se] = slptrain(net, xd, yd, opt)
% [net, se] = slptrain(net, xd, yd, opt) 
%   Funkcja uczy siec odpowiadac sygnalami yd na sygnaly wejscowe xd,
%   parametr opt decyduje o parametrach procesu uczenia
% 
% Argumenty wejsciowe:
%   net  - zmienna reprezentujaca perceptron jednowarstwowy
%   xd   - kolumnowy wektor sygnalow wejsciowych lub 
%          macierz ktorej kolumny zawieraja kolejne wektory sygnalow wejsciowych
%   yd   - kolumnowy wektor sygnalow wyjsciowych lub 
%          macierz ktorej kolumny zawieraja kolejne wektory sygnalow wyjsciowych
%   opt  - tablica komorkowa zawierajaca opcje:
%
%     view - parametr decydujacy o pokazywaniu szczegolow przebiegu procesu
%              uczenia, dopuszczalne wartosci:
%          0 - szczegoly nie sa pokazywane
%          1 - w kazdym kroku algorytmu pokazywane sa:
%              numer iteracji oraz wartosc bezwgledna sumy bledow dla wzorcow 
%              zawartych w xd (se)
%          2 - w kazdym kroku algorytmu wywolywana jest funkcja 
%              uzytkownika, ktorej nazwe nalezy podac w opcji viewFn
%          domyslnie view = 1
%     viewFn - parametr zawierajacy nazwe funkcji uzytkownika ktora jest 
%          odpowiedzialna za pokazywanie szczegolow procesu uczenia,
%          funkcja ta powinna miec nastepujacy naglowek:
%          function F(net, it, se)
%          parametry wejsciowe funkcji zawieraja:
%            net - aktualne dane perceptronu
%            it  - numer aktualnej iteracji
%            se  - wartosc bezwgledna sumy bledow dla wzorcow zawartych w xd
%     maxit - okresla maksymalna liczbe prezentacji wzorcow
%          domyslnie maxit = 1000
%     c     - wspolczynnik wykorzystywany w trakcie uczenia
%          domyslnie c = 0.1
%     eps - okresla wymagana maksymalna wartosc ktora musi osiagnac suma bledow
%           dla wzorcow wykorzystywanych w trakcie uczenia
%           domyslnie eps = 0.001
%          
% Argumenty wyjsciowe:
%   net  - zmienna reprezentujaca siec zawierajaca nowe wartosci wag i biasow
%   se   - wartosc bezwzgledna z sumy bledow dla prezentowanych wzorcow
%
% Przyklad
%
%   n = slp(2, 1, 'hardlims')
%       tworzy neuron o dwoch wejsciach z biasem z bipolarna funkcja skoku
%   x1 = [1;  1]; y1 = -1; 
%   x2 = [1; -1]; y2 =  1;
%       neuron powinien punkt (1,1) przypisywac do klasy (1)
%       a punkt (1, -1) do klasy (-1)
%   xd = [x1 x2]
%   yd = [y1 y2]
%       definicja wzorcow uczacych zgodnie z wymogami funkcji slptrain
%   n = slptrain(n, xd, yd)
%       zwraca siec ze zmodyfikowany w trakcie uczenia wagami
%
%   y = slpfwd(n, xd)
%       zwraca wartosci na wyjsciu neuronu dla uczonych wzorcow
%

if (strcmp(net.type, 'slp') == 0) 
  error('Siec nie jest perceptronem jednowarstwowym\n');
end

if (size(xd,1) ~= net.nin)
  error('Nieprawidlowy rozmiar wektora sygnalow wejsciowych\n');
end

if (size(yd,1) ~= net.nout)
  error('Nieprawidlowy rozmiar wektora sygnalow wyjsciowych\n');
end

if (size(xd,2) ~= size(yd,2))
  error('Liczba wzorcow w xd nie jest rowna liczbie wzorcow w yd\n');
end

if ( nargin == 3 )
  opt = {};
end

view   = findopt(opt, 'view', 1);
viewFn = findopt(opt, 'viewFn', '');
maxit  = findopt(opt, 'maxit', 1000);
c      = findopt(opt, 'c', 0.1);
eps    = findopt(opt, 'eps', 0.001);

if sum(strcmp(net.actfn, {'hardlims', 'tanh'})) ~= 0
  c = c/2;
end

if ((view == 2) & (length(viewFn) == 0))
  error('Niezdefiniowany parametr viewFn\n');
end

stop=1; i=0; j=0; p=0; se=0;

nxd = size(xd, 2);

while (stop > eps)

  i = i+1; j = j+1; 

  xdj = xd(:, j);
  if (~isempty(net.b))  
    xdj = [xdj; 1]; 
  end
  fi = net.wb * xdj;

  switch net.actfn
    case 'linear'    
      y = fi;
    case 'hardlims'
      idx=find(fi==0); fi(idx)=-0.1; y=sign(fi);
    case 'hardlim'
      idx=find(fi==0); fi(idx)=-0.1; y=(sign(fi)+1)*0.5;
    case 'tanh'
      y=tanh(fi);
    case 'logistic'
      y=1./(1+exp(-fi));
  end
  
  e = yd(:,j)-y; se=se+sum(abs(e));

  net.wb = net.wb+c*e*xdj';
  net.w = net.wb(:, 1:net.nin);
  if (~isempty(net.b))
    net.b = net.wb(:, net.nin+1);
  end  

  if (j==nxd) 
    switch view
      case 1
        disp(sprintf('%5d. se = %6.4f', p+1, se));
      case 2
        feval(viewFn, net, p+1, se);      
    end 
    stop = se; p = p+1; j = 0; se = 0;
    if ((p >= maxit) | (stop <= eps)); se=stop; stop = 0; end
  end

end